/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the Open-Xchange, Inc. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2004-2010 Open-Xchange, Inc.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.usm.ox_json.impl;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;
import com.openexchange.usm.ox_json.JSONResult;
import com.openexchange.usm.ox_json.JSONResultType;

public class OXJSONSupport {

    private static final String WARNING_CATEGORY = "13";

    private static final String CATEGORY_FIELD = "category";

    public static final String CATEGORY_RESOURCE_TEMPORARY_DOWN = "4";

    public static final String CATEGORY_SUBSYSTEM_DOWN = "5";

    public static final String CATEGORY_SOCKET_CONNECTION_CORRUPT = "6";

    public static final String CATEGORY_RESOURCE_FULL_OR_BUSY = "11";

    public static final String ERROR_CODE_OBJECT_CHANGED = "FLD-0001";

    public static final String ERROR_CODE_FOLDER_MUST_NOT_BE_DELETED = "FLD-0090";

    public static final String ERROR_FIELD = "error";

    public static final String ERROR_CODE_FIELD = "code";

    public static final String DATA_FIELD = "data";

    public static final String UNKNOWN_ERROR_CODE = "unknown";

    private static final Set<String> _RELOGIN_ERROR_CODES = new HashSet<String>();

    public static JSONResult parseResult(String content, JSONResultType expectedResultType) throws IOException {
        JSONTokener tokener = new JSONTokener(content);
        try {
            char c = tokener.nextClean();
            switch (c) {
            case '[':
            case '(':
                tokener.back();
                JSONArrayResult result = new JSONArrayResult(new JSONArray(tokener));
                if (tokener.nextClean() == 0) {
                    if (isValidResultType(expectedResultType, JSONResultType.JSONArray))
                        return result;
                    JSONObject errorDetails = new JSONObject();
                    try {
                        errorDetails.put(DATA_FIELD, result.getJSONArray());
                        errorDetails.put(ERROR_FIELD, "Invalid array response from server, expected " + expectedResultType);
                    } catch (JSONException ignored) {
                    }
                    return new JSONErrorResult(errorDetails);
                }
                break;
            case '{':
                tokener.back();
                JSONObject object = new JSONObject(tokener);
                if (tokener.nextClean() == 0)
                    return generateJSONResultForJSONObject(object, expectedResultType);
                break;
            default:
                break;
            }
        } catch (JSONException ignored) {
        }
        if (isValidResultType(expectedResultType, JSONResultType.String))
            return new JSONStringResult(content);
        JSONObject errorDetails = new JSONObject();
        try {
            errorDetails.put(DATA_FIELD, content);
            errorDetails.put(ERROR_FIELD, "Invalid string response from server, expected " + expectedResultType);
        } catch (JSONException ignored) {
        }
        return new JSONErrorResult(errorDetails);
    }

    /**
     * if JSONObject has a field 'error' and the category of this error is not "13" then return JSONErrorResult if a field 'data' is
     * available and is not empty then observe the content of this field; if starts with '[' or '(' return JSONArrayResult if starts with
     * '{' return JSONObjectResult
     * 
     * @param object
     * @return
     */
    public static JSONResult generateJSONResultForJSONObject(JSONObject object, JSONResultType expectedResultType) {
        if (isOXError(object))
            return new JSONErrorResult(object);
        if (isValidResultType(expectedResultType, JSONResultType.JSONObject))
            return new JSONObjectResult(object);
        JSONObject errorDetails = new JSONObject();
        try {
            errorDetails.put(DATA_FIELD, object);
            errorDetails.put(ERROR_FIELD, "Invalid object response from server, expected " + expectedResultType);
        } catch (JSONException ignored) {
            // can not happen
        }
        return new JSONErrorResult(errorDetails);
    }

    private static boolean isOXError(JSONObject object) {
        if (!object.has(ERROR_FIELD))
            return false;
        if (!object.has(CATEGORY_FIELD))
            return true;
        if (!WARNING_CATEGORY.equals(getJSONErrorCategory(object)))
            return true;
        return getJSONErrorCode(object).equals(ERROR_CODE_FOLDER_MUST_NOT_BE_DELETED);
    }

    public static String getJSONErrorCode(JSONObject object) {
        if (!object.has(ERROR_FIELD))
            return "";
        try {
            return object.getString(ERROR_CODE_FIELD);
        } catch (JSONException e) {
            return UNKNOWN_ERROR_CODE;
        }
    }

    public static String getJSONErrorCategory(JSONObject object) {
        return object.optString(CATEGORY_FIELD, null);
    }

    private static boolean isValidResultType(JSONResultType expected, JSONResultType actual) {
        return expected == null || expected == actual;
    }

    public static void setReloginErrorCodes(String reloginErrorCodes) {
        _RELOGIN_ERROR_CODES.clear();
        for (String s : reloginErrorCodes.split(",")) {
            String code = s.trim();
            if (code.length() > 0)
                _RELOGIN_ERROR_CODES.add(code);
        }
    }

    public static boolean shouldPerformNewLogin(JSONResult result) {
        return result.getResultType() == JSONResultType.Error && _RELOGIN_ERROR_CODES.contains(OXJSONSupport.getJSONErrorCode(result.getJSONObject()));
    }
}
